from enum import Enum
from croniter import croniter


class BatchSize(Enum):
    YEARLY = "YEARLY"
    MONTHLY = "MONTHLY"
    WEEKLY = "WEEKLY"
    DAILY = "DAILY"
    HOURLY = "HOURLY"
    SUBHOURLY = "SUBHOURLY"


def infer_batch_size(cron: str) -> BatchSize | None:
    if not croniter.is_valid(cron):
        raise ValueError(f"Invalid cron expression: {cron!r}")

    minute, hour, day, month, weekday = cron.split()

    if month != "*":
        return BatchSize.YEARLY
    if day != "*" and weekday == "*":
        return BatchSize.MONTHLY
    if weekday != "*":
        return BatchSize.WEEKLY
    if hour != "*":
        return BatchSize.DAILY
    if minute == "0":
        return BatchSize.HOURLY
    return None



from dataclasses import dataclass
from bollhav.database import DatabaseColumn
from enum import Enum


class PostgresType(Enum):
    # Numeric
    SMALLINT = "SMALLINT"
    INTEGER = "INTEGER"
    BIGINT = "BIGINT"
    DECIMAL = "DECIMAL"
    NUMERIC = "NUMERIC"
    REAL = "REAL"
    DOUBLE_PRECISION = "DOUBLE PRECISION"
    SMALLSERIAL = "SMALLSERIAL"
    SERIAL = "SERIAL"
    BIGSERIAL = "BIGSERIAL"
    # Monetary
    MONEY = "MONEY"
    # Character
    CHAR = "CHAR"
    VARCHAR = "VARCHAR"
    TEXT = "TEXT"
    # Binary
    BYTEA = "BYTEA"
    # Date/Time
    TIMESTAMP = "TIMESTAMP"
    TIMESTAMPTZ = "TIMESTAMPTZ"
    DATE = "DATE"
    TIME = "TIME"
    TIMETZ = "TIMETZ"
    INTERVAL = "INTERVAL"
    # Boolean
    BOOLEAN = "BOOLEAN"
    # Geometric
    POINT = "POINT"
    LINE = "LINE"
    LSEG = "LSEG"
    BOX = "BOX"
    PATH = "PATH"
    POLYGON = "POLYGON"
    CIRCLE = "CIRCLE"
    # Network
    CIDR = "CIDR"
    INET = "INET"
    MACADDR = "MACADDR"
    MACADDR8 = "MACADDR8"
    # Bit String
    BIT = "BIT"
    VARBIT = "VARBIT"
    # Text Search
    TSVECTOR = "TSVECTOR"
    TSQUERY = "TSQUERY"
    # UUID
    UUID = "UUID"
    # XML
    XML = "XML"
    # JSON
    JSON = "JSON"
    JSONB = "JSONB"
    # Arrays
    ARRAY = "ARRAY"
    # Range
    INT4RANGE = "INT4RANGE"
    INT8RANGE = "INT8RANGE"
    NUMRANGE = "NUMRANGE"
    TSRANGE = "TSRANGE"
    TSTZRANGE = "TSTZRANGE"
    DATERANGE = "DATERANGE"
    # Multirange (Postgres 14+)
    INT4MULTIRANGE = "INT4MULTIRANGE"
    INT8MULTIRANGE = "INT8MULTIRANGE"
    NUMMULTIRANGE = "NUMMULTIRANGE"
    TSMULTIRANGE = "TSMULTIRANGE"
    TSTZMULTIRANGE = "TSTZMULTIRANGE"
    DATEMULTIRANGE = "DATEMULTIRANGE"
    # Object Identifier
    OID = "OID"
    # Pseudo
    VOID = "VOID"


@dataclass
class PostgresColumn(DatabaseColumn):
    data_type: PostgresType = PostgresType.TEXT
    primary_key: bool = False
    unique: bool = False
    precision: int | None = None
    scale: int | None = None
    length: int | None = None

    def __post_init__(self) -> None:
        if self.primary_key and self.nullable:
            raise ValueError(
                f"Column {self.name!r}: primary_key=True cannot be nullable"
            )

    def __repr__(self) -> str:
        return (
            f"PostgresColumn(name={self.name!r}, data_type={self.data_type}, "
            f"nullable={self.nullable}, order={self.order}, primary_key={self.primary_key}, "
            f"unique={self.unique}, precision={self.precision}, scale={self.scale}, "
            f"length={self.length})"
        )



from enum import Enum


class WriteMode(Enum):
    APPEND = "APPEND"
    TRUNCATE_INSERT = "TRUNCATE_INSERT"
    OVERWRITE_INSERT = "OVERWRITE_INSERT"
    MERGE = "MERGE"
    VIEW = "VIEW"


class ModelType(Enum):
    TABLE = "TABLE"
    VIEW = "VIEW"



import pytest
import polars as pl

from bollhav import Model, ModelType, WriteMode, BatchSize
from bollhav.batching import infer_batch_size
from bollhav.database import Database, DatabaseColumn
from bollhav.postgres import PostgresType, PostgresColumn


COLUMNS: dict[str, pl.DataType] = {"id": pl.Int64, "amount": pl.Float64}


class TestModelDefaults:
    def test_defaults(self):
        m = Model(name="orders", source_table="raw.orders", columns=COLUMNS)
        assert m.model_type == ModelType.TABLE
        assert m.write_mode == WriteMode.APPEND
        assert m.tags is None
        assert m.cron is None
        assert m.batch_size is None
        assert m.extra == {}

    def test_destination_defaults_to_empty(self):
        m = Model(name="orders", source_table="raw.orders", columns=COLUMNS)
        assert m.destination_table == ""
        assert m.destination_schema == ""

    def test_destination_db_defaults_to_none(self):
        m = Model(name="orders", source_table="raw.orders", columns=COLUMNS)
        assert m.destination_db is None
        assert m.destination_columns is None


class TestModelValidation:
    def test_view_type_requires_view_write_mode(self):
        with pytest.raises(ValueError):
            Model(
                name="v",
                source_table="raw.orders",
                columns=COLUMNS,
                model_type=ModelType.VIEW,
                write_mode=WriteMode.APPEND,
            )

    def test_table_type_cannot_use_view_write_mode(self):
        with pytest.raises(ValueError):
            Model(
                name="t",
                source_table="raw.orders",
                columns=COLUMNS,
                model_type=ModelType.TABLE,
                write_mode=WriteMode.VIEW,
            )

    def test_view_type_with_view_write_mode_is_valid(self):
        m = Model(
            name="v",
            source_table="raw.orders",
            columns=COLUMNS,
            model_type=ModelType.VIEW,
            write_mode=WriteMode.VIEW,
        )
        assert m.model_type == ModelType.VIEW
        assert m.write_mode == WriteMode.VIEW

    def test_destination_db_without_columns_raises(self):
        with pytest.raises(ValueError):
            Model(
                name="orders",
                source_table="raw.orders",
                columns=COLUMNS,
                destination_db=Database.POSTGRES,
            )

    def test_destination_columns_without_db_raises(self):
        with pytest.raises(ValueError):
            Model(
                name="orders",
                source_table="raw.orders",
                columns=COLUMNS,
                destination_columns=[
                    PostgresColumn(name="id", data_type=PostgresType.INTEGER)
                ],
            )

    def test_destination_db_and_columns_valid(self):
        m = Model(
            name="orders",
            source_table="raw.orders",
            columns=COLUMNS,
            destination_db=Database.POSTGRES,
            destination_columns=[
                PostgresColumn(name="id", data_type=PostgresType.INTEGER)
            ],
        )
        assert m.destination_db == Database.POSTGRES
        assert len(m.destination_columns) == 1


class TestBatchSizeInference:
    @pytest.mark.parametrize(
        "cron,expected",
        [
            ("0 0 1 1 *", BatchSize.YEARLY),
            ("0 0 1 * *", BatchSize.MONTHLY),
            ("0 0 * * 0", BatchSize.WEEKLY),
            ("0 3 * * *", BatchSize.DAILY),
            ("0 * * * *", BatchSize.HOURLY),
            ("*/15 * * * *", None),
        ],
    )
    def test_infer_batch_size(self, cron: str, expected: BatchSize):
        assert infer_batch_size(cron) == expected

    def test_invalid_cron_raises(self):
        with pytest.raises(ValueError):
            infer_batch_size("not a cron")

    def test_batch_size_set_on_model(self):
        m = Model(
            name="orders", source_table="raw.orders", columns=COLUMNS, cron="0 3 * * *"
        )
        assert m.batch_size == BatchSize.DAILY

    def test_no_cron_gives_none_batch_size(self):
        m = Model(name="orders", source_table="raw.orders", columns=COLUMNS)
        assert m.batch_size is None


class TestPostgresColumn:
    def test_defaults(self):
        col = PostgresColumn(name="id", data_type=PostgresType.INTEGER)
        assert col.nullable is True
        assert col.order is None
        assert col.primary_key is False
        assert col.unique is False

    def test_primary_key_nullable_raises(self):
        with pytest.raises(ValueError, match="primary_key=True cannot be nullable"):
            PostgresColumn(
                name="id",
                data_type=PostgresType.INTEGER,
                primary_key=True,
                nullable=True,
            )

    def test_primary_key_not_nullable_valid(self):
        col = PostgresColumn(
            name="id", data_type=PostgresType.INTEGER, primary_key=True, nullable=False
        )
        assert col.primary_key is True
        assert col.nullable is False

    def test_unique(self):
        col = PostgresColumn(name="email", data_type=PostgresType.TEXT, unique=True)
        assert col.unique is True

    def test_order(self):
        col = PostgresColumn(
            name="created_at", data_type=PostgresType.TIMESTAMPTZ, order=2
        )
        assert col.order == 2

    def test_inherits_database_column(self):
        col = PostgresColumn(name="id", data_type=PostgresType.INTEGER)
        assert isinstance(col, DatabaseColumn)

    def test_repr(self):
        col = PostgresColumn(
            name="id", data_type=PostgresType.INTEGER, nullable=False, primary_key=True
        )
        assert "id" in repr(col)
        assert "INTEGER" in repr(col)
        assert "primary_key=True" in repr(col)

    def test_equality(self):
        a = PostgresColumn(name="id", data_type=PostgresType.INTEGER)
        b = PostgresColumn(name="id", data_type=PostgresType.INTEGER)
        assert a == b

    def test_inequality(self):
        a = PostgresColumn(name="id", data_type=PostgresType.INTEGER)
        b = PostgresColumn(name="id", data_type=PostgresType.BIGINT)
        assert a != b


class TestExtra:
    def test_plain_kwargs_stored_in_extra(self):
        m = Model(name="orders", source_table="raw.orders", columns=COLUMNS, foo="bar")
        assert m.extra["foo"] == "bar"

    def test_callable_kwargs_resolved(self):
        def my_ddl(table_name: str, **kwargs) -> str:
            return f"CREATE TABLE {table_name};"

        m = Model(
            name="orders",
            source_table="raw.orders",
            columns=COLUMNS,
            destination_ddl=my_ddl,
            table_name="orders",
        )
        assert m.extra["destination_ddl"] == "CREATE TABLE orders;"

    def test_callable_receives_non_callable_kwargs(self):
        def my_indexes(table_name: str, schema: str, **kwargs) -> list[str]:
            return [f"CREATE INDEX ON {schema}.{table_name} (id)"]

        m = Model(
            name="orders",
            source_table="raw.orders",
            columns=COLUMNS,
            destination_indexes=my_indexes,
            table_name="orders",
            schema="public",
        )
        assert m.extra["destination_indexes"] == ["CREATE INDEX ON public.orders (id)"]


class TestEquality:
    def test_equal_models(self):
        a = Model(name="orders", source_table="raw.orders", columns=COLUMNS)
        b = Model(name="orders", source_table="raw.orders", columns=COLUMNS)
        assert a == b

    def test_unequal_models(self):
        a = Model(name="orders", source_table="raw.orders", columns=COLUMNS)
        b = Model(name="other", source_table="raw.orders", columns=COLUMNS)
        assert a != b

    def test_not_equal_to_non_model(self):
        m = Model(name="orders", source_table="raw.orders", columns=COLUMNS)
        assert m.__eq__("not a model") == NotImplemented


class TestRepr:
    def test_repr_contains_name(self):
        m = Model(name="orders", source_table="raw.orders", columns=COLUMNS)
        assert "orders" in repr(m)

    def test_repr_contains_destination_db(self):
        m = Model(
            name="orders",
            source_table="raw.orders",
            columns=COLUMNS,
            destination_db=Database.POSTGRES,
            destination_columns=[
                PostgresColumn(name="id", data_type=PostgresType.INTEGER)
            ],
        )
        assert "POSTGRES" in repr(m)



import polars as pl
from enum import Enum
from croniter import croniter
from bollhav.database import Database
from bollhav.postgres import PostgresColumn
from bollhav.modes import WriteMode, ModelType
from bollhav.batching import BatchSize


def _infer_batch_size(cron: str) -> BatchSize | None:
    if not croniter.is_valid(cron):
        raise ValueError(f"Invalid cron expression: {cron!r}")

    minute, hour, day, month, weekday = cron.split()

    if month != "*":
        return BatchSize.YEARLY
    if day != "*" and weekday == "*":
        return BatchSize.MONTHLY
    if weekday != "*":
        return BatchSize.WEEKLY
    if hour != "*":
        return BatchSize.DAILY
    if minute == "0":
        return BatchSize.HOURLY
    return None


class Model:
    def __init__(
        self,
        name: str,
        source_table: str,
        columns: dict[str, pl.DataType],
        destination_table: str = "",
        destination_schema: str = "",
        destination_db: Database | None = None,
        destination_columns: list[PostgresColumn] | None = None,
        model_type: ModelType = ModelType.TABLE,
        write_mode: WriteMode = WriteMode.APPEND,
        tags: list[str] | None = None,
        cron: str | None = None,
        **kwargs,
    ):
        if model_type == ModelType.VIEW and write_mode != WriteMode.VIEW:
            raise ValueError(f"ModelType.VIEW must use WriteMode.VIEW")
        if model_type == ModelType.TABLE and write_mode == WriteMode.VIEW:
            raise ValueError(f"ModelType.TABLE cannot use WriteMode.VIEW")
        if destination_db is not None and destination_columns is None:
            raise ValueError(
                "destination_columns must be set when destination_db is provided"
            )
        if destination_columns is not None and destination_db is None:
            raise ValueError(
                "destination_db must be set when destination_columns is provided"
            )

        self.name = name
        self.source_table = source_table
        self.destination_table = destination_table
        self.destination_schema = destination_schema
        self.destination_db = destination_db
        self.destination_columns = destination_columns
        self.model_type = model_type
        self.write_mode = write_mode
        self.tags = tags
        self.columns = columns
        self.cron = cron
        self.batch_size = _infer_batch_size(cron) if cron else None

        for key, val in kwargs.items():
            if callable(val):
                kwargs[key] = val(
                    **{k: v for k, v in kwargs.items() if not callable(v)}
                )
        self.extra = kwargs

    def __repr__(self) -> str:
        return (
            f"Model(name={self.name!r}, source_table={self.source_table!r}, "
            f"destination_table={self.destination_table!r}, destination_schema={self.destination_schema!r}, "
            f"destination_db={self.destination_db}, destination_columns={self.destination_columns!r}, "
            f"model_type={self.model_type}, write_mode={self.write_mode}, "
            f"tags={self.tags!r}, columns={self.columns!r}, cron={self.cron!r}, "
            f"batch_size={self.batch_size}, extra={self.extra!r})"
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Model):
            return NotImplemented
        return self.__dict__ == other.__dict__
